import collections
from pprint import pformat

from potion_client.utils import escape


class PaginatedList(collections.abc.Sequence):
    def __init__(self, binding, params):
        self._pages = {}
        self._per_page = per_page = params.pop("per_page", 20)
        self._binding = binding
        self._total_count = 0
        self._request_params = params
        self.fetch_page(1, per_page)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [self.__getitem__(index) for index in range(*item.indices(self._total_count))]

        if item < 0 or item >= self._total_count:
            raise IndexError()

        page, offset = item // self._per_page + 1, item % self._per_page
        if page not in self._pages:
            self.fetch_page(page, self._per_page)
        return self._pages[page][offset]

    def __len__(self):
        return self._total_count

    def fetch_page(self, page, per_page):
        params = {"page": page, "per_page": per_page}
        params.update(self._request_params)
        response, response_data = self._binding.make_request(None, params)

        try:
            self._total_count = int(response.headers["X-Total-Count"])
        except KeyError:
            self._total_count = len(response_data)

        self._pages[page] = response_data

    def _repr_html_(self):
        if len(self) <= 10:
            items = [escape(pformat(item)) for item in self[:]]
        else:
            items = (
                [escape(pformat(item)) for item in self[:5]]
                + ["<strong>...</strong>"]
                + [escape(pformat(item)) for item in self[-5:]]
            )

        return """<table>
        <thead>
            <tr>
                <th><code>PaginatedList({params})</code>: <em>{count}</em> items</th>
            </tr>
        </thead>
        <tbody>{items}</tbody>
        </table>""".format(
            params=", ".join(
                ["{}.{}".format(self._binding.owner.__name__, self._binding.link.rel)]
                + ["{}={}".format(k, repr(v)) for k, v in self._request_params.items()]
            ),
            count=self._total_count,
            items="\n".join("<tr><td><code>{}</code></td></tr>".format(item) for item in items),
        )

    def __repr__(self):
        return "PaginatedList({params})".format(
            params=", ".join(
                ["{}.{}".format(self._binding.owner.__name__, self._binding.link.rel)]
                + ["{}={}".format(k, repr(v)) for k, v in self._request_params.items()]
            )
        )



import collections
import re


class Schema(collections.abc.Mapping):
    def __init__(self, schema):
        if isinstance(schema, Schema):
            schema = schema._schema
        self._schema = schema or {}

    @property
    def type(self):
        try:
            type = self._schema["type"]

            if isinstance(type, (list, tuple)):
                return tuple(type)

            return (type,)
        except KeyError:
            return None

    @property
    def readonly_properties(self):
        if self.type != "object":
            return ()

        properties = []
        for name, schema in self._schema.get("properties", {}).items():
            if schema.get("readOnly", False):
                properties.append(name)

        return tuple(properties)

    @property
    def required_properties(self):
        if "object" not in self.type:
            return ()

        return tuple(self._schema.get("required", []))

    def can_include_property(self, name):
        # empty schema "{}" allows all properties
        if not self._schema:
            return True

        # only objects can have properties
        if "object" not in self.type:
            return False

        try:
            property_schema = self._schema["properties"][name]

            if property_schema.get("readOnly", False):
                return False

            return True
        except KeyError:
            if self._schema.get("additionalProperties", True):
                return True

            for pattern in self._schema.get("patternProperties", {}).values():
                if re.match(pattern, name):
                    return True

            return False

    def __contains__(self, item):
        return item in self._schema

    def __getitem__(self, item):
        return self._schema[item]

    def __iter__(self):
        return iter(self._schema)

    def __len__(self):
        return len(self._schema)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self._schema))



import collections
from functools import partial
from operator import delitem, getitem, setitem
from weakref import WeakValueDictionary

import requests
from six.moves.urllib.parse import urljoin, urlparse

from potion_client.converter import PotionJSONDecoder, PotionJSONSchemaDecoder
from potion_client.links import Link
from potion_client.resource import Reference, Resource, uri_for
from potion_client.utils import snake_case, upper_camel_case

__version__ = "2.5.4"


class Client:
    # TODO optional HTTP/2 support: this makes multiple queries simultaneously.

    def __init__(self, api_root_url, schema_path="/schema", fetch_schema=True, **session_kwargs):
        self._instances = WeakValueDictionary()
        self._resources = {}

        self.session = session = requests.Session()
        for key, value in session_kwargs.items():
            setattr(session, key, value)

        parse_result = urlparse(api_root_url)
        self._root_url = "{}://{}".format(parse_result.scheme, parse_result.netloc)
        self._api_root_url = api_root_url  # '{}://{}'.format(parse_result.scheme, parse_result.netloc)
        self._root_path = parse_result.path
        self._schema_url = api_root_url + schema_path

        if fetch_schema:
            self._fetch_schema()

    def _fetch_schema(self):
        schema = self.session.get(self._schema_url).json(
            cls=PotionJSONSchemaDecoder, referrer=self._schema_url, client=self
        )

        # NOTE these should perhaps be definitions in Flask-Potion
        for name, resource_schema in schema["properties"].items():
            resource = self.resource_factory(name, resource_schema)
            setattr(self, upper_camel_case(name), resource)

    def instance(self, uri, cls=None, default=None, **kwargs):
        instance = self._instances.get(uri, None)

        if instance is None:
            if cls is None:
                try:
                    cls = self._resources[uri[: uri.rfind("/")]]
                except KeyError:
                    cls = Reference

            if isinstance(default, Resource) and default._uri is None:
                default._status = 200
                default._uri = uri
                instance = default
            else:
                instance = cls(uri=uri, **kwargs)
            self._instances[uri] = instance
        return instance

    def fetch(self, uri, cls=PotionJSONDecoder, **kwargs):
        # TODO handle URL fragments (#properties/id etc.)
        response = self.session.get(urljoin(self._root_url, uri, True))

        response.raise_for_status()

        return response.json(cls=cls, client=self, referrer=uri, **kwargs)

    def resource_factory(self, name, schema, resource_cls=None):
        """
        Registers a new resource with a given schema. The schema must not have any unresolved references
        (such as `{"$ref": "#"}` for self-references, or otherwise). A subclass of :class:`Resource`
        may be provided to add specific functionality to the resulting :class:`Resource`.

        :param str name:
        :param dict schema:
        :param Resource resource_cls: a subclass of :class:`Resource` or None
        :return: The new :class:`Resource`.
        """
        cls = type(
            str(upper_camel_case(name)),
            (resource_cls or Resource, collections.abc.MutableMapping),
            {"__doc__": schema.get("description", "")},
        )

        cls._schema = schema
        cls._client = self
        cls._links = links = {}

        for link_schema in schema["links"]:
            link = Link(
                self,
                rel=link_schema["rel"],
                href=link_schema["href"],
                method=link_schema["method"],
                schema=link_schema.get("schema", None),
                target_schema=link_schema.get("targetSchema", None),
            )

            # Set Resource._self, etc. for the special methods as they are managed by the Resource class
            if link.rel in ("self", "instances", "create", "update", "destroy"):
                setattr(cls, "_{}".format(link.rel), link)
            links[link.rel] = link

            if link.rel != "update":  # 'update' is a special case because of MutableMapping.update()
                setattr(cls, snake_case(link.rel), link)

        # TODO routes (instance & non-instance)

        for property_name, property_schema in schema.get("properties", {}).items():
            # skip $uri and $id as these are already implemented in Resource and overriding them causes unnecessary
            # fetches.
            if property_name.startswith("$"):
                continue

            if property_schema.get("readOnly", False):
                # TODO better error message. Raises AttributeError("can't set attribute")
                setattr(
                    cls,
                    property_name,
                    property(
                        fget=partial((lambda name, obj: getitem(obj, name)), property_name),
                        doc=property_schema.get("description", None),
                    ),
                )
            else:
                setattr(
                    cls,
                    property_name,
                    property(
                        fget=partial((lambda name, obj: getitem(obj, name)), property_name),
                        fset=partial((lambda name, obj, value: setitem(obj, name, value)), property_name),
                        fdel=partial((lambda name, obj: delitem(obj, name)), property_name),
                        doc=property_schema.get("description", None),
                    ),
                )

        root = None
        if "instances" in links:
            root = cls._instances.href
        elif "self" in links:
            root = cls._self.href[: cls._self.href.rfind("/")]
        else:
            root = self._root_path + "/" + name.replace("_", "-")

        self._resources[root] = cls
        return cls


ASC = ASCENDING = False
DESC = DESCENDING = True



import calendar
from datetime import date, datetime
from functools import partial
from json import JSONDecoder, JSONEncoder

from six.moves.urllib.parse import urljoin

from potion_client.resource import Reference

try:
    from datetime import timezone
except ImportError:
    from datetime import tzinfo, timedelta

    # pragma: no cover
    class timezone(tzinfo):
        def __init__(self, utcoffset, name=None):
            self._utcoffset = utcoffset
            self._name = name

        def utcoffset(self, dt):
            return self._utcoffset

        def tzname(self, dt):
            return self._name

        def dst(self, dt):
            return timedelta(0)

    timezone.utc = timezone(timedelta(0), "UTC")


class PotionJSONEncoder(JSONEncoder):
    def encode(self, o):
        root_id = id(o)
        if self.check_circular:
            markers = {}
        else:
            markers = None

        def _encode(o):
            if isinstance(o, (list, tuple, dict)):
                if markers is not None:
                    marker_id = id(o)
                    if marker_id in markers:
                        if marker_id == root_id:
                            return {"$ref": "#"}
                        raise ValueError("Circular reference detected")
                    markers[marker_id] = o
                try:
                    if isinstance(o, dict):
                        return {k: _encode(v) for k, v in o.items()}
                    else:
                        return [_encode(v) for v in o]
                finally:
                    if markers is not None:
                        del markers[marker_id]

            if isinstance(o, date):
                return {"$date": int(calendar.timegm(o.timetuple()) * 1000)}
            if isinstance(o, datetime):
                return {"$date": int(calendar.timegm(o.utctimetuple()) * 1000)}

            if isinstance(o, Reference):
                # FIXME if reference is not saved, save it first here
                return {"$ref": o._uri}

            return o

        return JSONEncoder.encode(self, _encode(o))


class PotionJSONDecoder(JSONDecoder):
    def __init__(self, client, referrer=None, uri_to_instance=True, default_instance=None, *args, **kwargs):
        self.client = client
        self.referrer = referrer
        self.uri_to_instance = uri_to_instance
        self.default_instance = default_instance
        JSONDecoder.__init__(self, *args, **kwargs)

    def _decode(self, o, depth=0):
        if isinstance(o, dict):
            if len(o) == 1:
                if "$date" in o:
                    return datetime.fromtimestamp(o["$date"] / 1000.0, timezone.utc)
                if "$ref" in o and isinstance(o["$ref"], str):
                    reference = o["$ref"]
                    if reference.startswith("#"):
                        reference = urljoin(self.referrer, reference, True)
                    return self.client.instance(reference)
            elif self.uri_to_instance and "$uri" in o and isinstance(o["$uri"], str):
                # TODO handle or ("$id" in o and "$type" in o)
                if depth == 0:
                    instance = self.client.instance(o["$uri"], default=self.default_instance)
                else:
                    instance = self.client.instance(o["$uri"])

                instance._status = 200
                instance._properties.update({k: self._decode(v, depth + 1) for k, v in o.items()})
                return instance

            return {k: self._decode(v, depth + 1) for k, v in o.items()}
        if isinstance(o, (list, tuple)):
            return [self._decode(v, depth + 1) for v in o]
        return o

    def decode(self, s, *args, **kwargs):
        o = JSONDecoder.decode(self, s, *args, **kwargs)
        return self._decode(o)


class JSONSchemaReference(Reference):
    @classmethod
    def _resolve(self, client, uri):
        return client.fetch(uri, cls=PotionJSONSchemaDecoder)


class PotionJSONSchemaDecoder(JSONDecoder):
    def __init__(self, client, referrer=None, *args, **kwargs):
        self.client = client
        self.referrer = referrer
        JSONDecoder.__init__(self, *args, **kwargs)

    def decode(self, s, *args, **kwargs):
        o = JSONDecoder.decode(self, s, *args, **kwargs)
        return schema_resolve_refs(o, partial(self.client.instance, cls=JSONSchemaReference, client=self.client))


def schema_resolve_refs(schema, ref_resolver=None, root=None):
    """
    Helper method for decoding references. Self-references are resolved automatically; other references are
    resolved using a callback function.

    :param object schema:
    :param callable ref_resolver:
    :param None root:
    :return:
    """
    # FIXME more stable implementation that only attempts to resolve {"$ref"} objects where they are allowed.
    if isinstance(schema, dict):
        if len(schema) == 1 and "$ref" in schema and isinstance(schema["$ref"], str):
            reference = schema["$ref"]
            if reference.startswith("#"):
                # TODO should also resolve any paths within the reference, which would need to be deferred.
                return root
            return ref_resolver(reference)

        resolved = {}
        for k, v in schema.items():
            resolved[k] = schema_resolve_refs(v, ref_resolver=ref_resolver, root=root if root is not None else resolved)
        return resolved
    if isinstance(schema, (list, tuple)):
        return [schema_resolve_refs(v, ref_resolver=ref_resolver, root=root) for v in schema]
    return schema



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': '.'}

packages = \
['potion_client']

package_data = \
{'': ['*']}

install_requires = \
['jsonschema>=3.0,<4.0', 'requests>=2.5,<3.0', 'six>=1.12,<2.0']

setup_kwargs = {
    'name': 'turmalin',
    'version': '2.5.4',
    'description': 'A client for APIs written in Flask-Potion',
    'long_description': None,
    'author': 'Lars Schöning',
    'author_email': 'lars@lyschoening.de',
    'url': 'https://github.com/biosustain/potion-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import collections
from pprint import pformat

from potion_client.exceptions import ItemNotFound, MultipleItemsFound
from potion_client.utils import escape


def uri_for(reference):
    """
    Utility function for accessing the URI or a reference.

    :param reference:
    :return:
    """
    return reference._uri


class Reference(collections.abc.Mapping):
    """

    This implementation makes the assumption that a {$ref} object always points to an object, never an array or
    any of the other types.
    """

    _client = None
    __properties = None

    def __init__(self, uri, client=None):
        self._status = None
        self._uri = uri
        self.__properties = {"$uri": uri}
        if client is not None:
            self._client = client

    @classmethod
    def _resolve(self, client, uri):
        return client.fetch(uri, uri_to_instance=False)

    @property
    def _properties(self):
        if self._uri and self._status is None:
            self.__properties = self._resolve(self._client, self._uri)
            self._status = 200
        return self.__properties

    @_properties.setter
    def _properties(self, value):
        self.__properties = value
        self._status = 200

    def __contains__(self, item):
        return item in self._properties

    def __getitem__(self, item):
        return self._properties[item]

    def __iter__(self):
        return iter(self._properties)

    def __len__(self):
        return len(self._properties)

    def __repr__(self):
        return "{cls}({uri})".format(cls=self.__class__.__name__, uri=repr(self._uri))


class Resource(Reference):
    _client = None
    _links = None
    _self = None
    _instances = None
    _create = None
    _destroy = None
    _update = None

    def __new__(cls, uri=None, **kwargs):
        instance = None
        if uri is not None:
            if not (isinstance(uri, str) and uri.startswith("/")) and cls._self is not None:
                uri = cls._self.href.format(id=uri)

            instances = cls._client._instances
            instance = instances.get(uri, None)

        if instance is None:
            instance = super().__new__(cls)
            super(Resource, instance).__init__(uri)
            instance._properties = {"$uri": uri}
            if not kwargs:
                instance._status = None
            else:
                instance._status = 200
                instance._properties.update(kwargs)

            if uri is not None:
                instances[uri] = instance

        # NOTE ensures that there is a single instance of a Resource with a given URL unless one creates an item
        # without URL and creates an item with the URL the first item is going to have, before saving the first item.
        return instance

    def __init__(self, uri=None, **kwargs):
        pass  # Must be blank. See __new__()

    @property
    def id(self):
        if self._uri is not None:
            id_ = self._uri[self._uri.rfind("/") + 1 :]
            if id_.isdigit():
                return int(id_)
            return id_
        return None

    # TODO cache this property
    @property
    def _validator(self):
        return None

    def __delitem__(self, item):
        del self._properties[item]

    def __setitem__(self, item, value):
        self._properties[item] = value

    def update(self, *args, **kwargs):
        self._properties.update(*args, **kwargs)
        self.save()

    @classmethod
    def first(cls, **params):
        try:
            return cls._instances(per_page=1, **params)[0]
        except IndexError:
            raise ItemNotFound("No '{}' item found matching: {}".format(cls.__name__, repr(params)))

    @classmethod
    def one(cls, **params):
        matching = cls._instances(per_page=1, **params)
        if len(matching) > 1:
            raise MultipleItemsFound("Multiple items found matching: {}".format(repr(params)))
        try:
            return matching[0]
        except IndexError:
            raise ItemNotFound("No '{}' item found matching: {}".format(cls.__name__, repr(params)))

    @classmethod
    def fetch(cls, id):
        return cls._self(id=id)

    def refresh(self):
        return self._self(id=self.id)

    def check(self):
        pass

    def save(self):
        # TODO only save the appropriate properties defined in the schemas
        # That logic should live in the links themselves.

        if self._uri is None:
            return self._create(**self)
        else:
            return self._update(**self)

    def delete(self):
        return self._destroy(id=self.id)

    def _repr_html_(self):
        return """<table>
        <thead>
            <tr>
                <th colspan="2"><code>{cls}({id})</code></th>
            </tr>
        </thead>
        <tbody>{properties}</tbody>
        </table>""".format(
            cls=self.__class__.__name__,
            id=escape(repr(self.id)),
            properties="\n".join(
                "<tr><td>{}</td><td><code>{}</code></td>".format(escape(k), escape(pformat(v)))
                for k, v in self._properties.items()
                if not k.startswith("$")
            ),
        )

    def __repr__(self):
        # if self._status == 200:
        #     parts = ['{}={}'.format(k, repr(v)) for k, v in self._properties.items() if not k.startswith('$')]
        #     if parts:
        #         return '{cls}({id}, {properties})'.format(cls=self.__class__.__name__,
        #                                                   id=repr(self.id),
        #                                                   properties=', '.join(parts))
        # TODO some way to define a good key for display
        return "{cls}({id})".format(cls=self.__class__.__name__, id=repr(self.id))



import json
import re

from requests import Request
from requests.exceptions import HTTPError

from potion_client import PotionJSONDecoder
from potion_client.collection import PaginatedList
from potion_client.converter import PotionJSONEncoder
from potion_client.schema import Schema


class Link:
    def __init__(self, client, method, href, rel, schema=None, target_schema=None):
        self.method = method
        self.href_placeholders = re.findall(r"{(\w+)}", href)
        self.href = href
        self.rel = rel
        self.schema = Schema(schema)
        self.target_schema = Schema(target_schema)

    @property
    def requires_instance(self):
        return "{id}" in self.href

    def returns_pagination(self):
        if self.method == "GET" and self.schema is not None:
            schema_properties = self.schema.get("properties", {})
            return "page" in schema_properties and "per_page" in schema_properties
        return False

    def __get__(self, instance, owner):
        return LinkBinding(self, instance, owner)


class LinkBinding:
    def __init__(self, link, instance, owner):
        self.link = link
        self.instance = instance
        self.owner = owner

    def request_factory(self, data, params):
        if self.instance is None:
            request_url = self.owner._client._root_url + self.link.href.format(**params)
        else:
            request_url = self.owner._client._root_url + self.link.href.format(id=self.instance.id, **self.instance)

        request_data = data
        request_params = {
            name: value
            for name, value in params.items()
            if name not in self.link.href_placeholders and self.link.schema.can_include_property(name)
        }

        if data is None:
            request_data = request_params
        elif isinstance(data, dict):
            request_params = data

        if self.link.method == "GET":
            req = Request(
                self.link.method,
                request_url,
                params={k: json.dumps(v, cls=PotionJSONEncoder) for k, v in request_params.items()},
            )
        else:
            req = Request(
                self.link.method,
                request_url,
                headers={"content-type": "application/json"},
                data=json.dumps(request_data, cls=PotionJSONEncoder),
            )
        return req

    def raise_for_status(self, response):
        http_error_msg = ""

        if 400 <= response.status_code < 500:
            try:
                http_error_msg = response.json()
            except Exception:
                http_error_msg = "{code} Client Error: {reason} for url: {url}".format(
                    code=response.status_code, reason=response.reason, url=response.url
                )

        elif 500 <= response.status_code < 600:
            http_error_msg = "{code} Server Error: {reason} for url: {url}".format(
                code=response.status_code, reason=response.reason, url=response.url
            )

        if http_error_msg:
            raise HTTPError(http_error_msg, response=response)

    def make_request(self, data, params):
        req = self.request_factory(data, params)
        prepared_request = self.owner._client.session.prepare_request(req)

        response = self.owner._client.session.send(prepared_request)

        # return error for some error conditions
        self.raise_for_status(response)

        if response.status_code == 204:
            return response, None

        return response, response.json(cls=PotionJSONDecoder, client=self.owner._client, default_instance=self.instance)

    def __getattr__(self, item):
        return getattr(self.link, item)

    def __call__(self, *arg, **params):
        data = None

        # Need to pass positional argument as *arg so that properties of the same name are not overridden in **params.
        if len(arg) > 1:
            raise TypeError("Link must be called with no more than one positional argument")
        elif len(arg) == 1:
            data = arg[0]

        if self.link.returns_pagination():
            return PaginatedList(self, params)

        response, response_data = self.make_request(data, params)
        return response_data



import re

__author__ = "lyschoening"


def camel_case(s):
    return s[0].lower() + s.title().replace("-", "").replace("_", "")[1:] if s else s


def upper_camel_case(s):
    return s.title().replace("-", "").replace("_", "") if s else s


def snake_case(s):
    return s[0].lower() + re.sub("([A-Z])", r"_\1", s[1:]).lower()


def escape(html):
    """Returns the given HTML with ampersands, quotes and carets encoded."""
    return (
        html.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
        .replace("'", "&#39;")
    )
